#include <Arduino.h>
#include <TM1637Display.h>
#include <Ticker.h>
#include <DNSServer.h>
#ifdef ESP32
#include <WiFi.h>
#include <AsyncTCP.h>
#elif defined(ESP8266)
#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#endif
#include "ESPAsyncWebServer.h"
#include "LittleFS.h"

// Module connection pins (Digital Pins)
#define CLK 2
#define DIO_GOF_7SEG 0
#define DIO_MIL_7SEG 5
#define DIO_KGG_7SEG 4

#define DIO_GOF_TRG 12
#define DIO_MIL_TRG 13
#define DIO_KGG_TRG 14

enum Parties
{
  NONE,
  GOF,
  MILIZ,
  KGG
};

void SevenSeg_Output();
void ticker_callback();
void serialOut_callback();
void inputGetter_callback();

TM1637Display disp_GOF(CLK, DIO_GOF_7SEG);
TM1637Display disp_MIL(CLK, DIO_MIL_7SEG);
TM1637Display disp_KGG(CLK, DIO_KGG_7SEG);

DNSServer dnsServer;
AsyncWebServer server(80);

Ticker PartyTicker(ticker_callback, 500, 0, MILLIS);
Ticker SerialOutputTicker(serialOut_callback, 5000, 0, MILLIS);
Ticker InputGetterTicker(inputGetter_callback, 500, 0, MILLIS);

Parties activeParty = NONE;

uint32_t Count_GOF = 0;
uint32_t Count_MILIZ = 0;
uint32_t Count_KGG = 0;
uint8_t KGG_dot = 0;
uint8_t GOF_dot = 0;
uint8_t MIL_dot = 0;

String processor(const String &var)
{
  char buffer[16] = {0};

  if (var == "POINTS_KGG")
    itoa(Count_KGG, buffer, 10);

  if (var == "POINTS_GOF")
    itoa(Count_GOF, buffer, 10);

  if (var == "POINTS_MILIZ")
    itoa(Count_MILIZ, buffer, 10);

  if (var == "ACTIVE_KGG")
    return activeParty == KGG ? "ACTIVE" : "INACTIVE";

  if (var == "ACTIVE_GOF")
    return activeParty == GOF ? "ACTIVE" : "INACTIVE";

  if (var == "ACTIVE_MILIZ")
    return activeParty == MILIZ ? "ACTIVE" : "INACTIVE";

  return String(buffer);
}

String getValue(String data, char separator, int index)
{
  int found = 0;
  int strIndex[] = {0, -1};
  int maxIndex = data.length() - 1;

  for (int i = 0; i <= maxIndex && found <= index; i++)
  {
    if (data.charAt(i) == separator || i == maxIndex)
    {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i + 1 : i;
    }
  }

  return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}

class CaptiveRequestHandler : public AsyncWebHandler
{
public:
  CaptiveRequestHandler() {}
  virtual ~CaptiveRequestHandler() {}

  bool canHandle(AsyncWebServerRequest *request)
  {
    return true;
  }

  void handleRequest(AsyncWebServerRequest *request)
  {

    Serial.print("Requested URL: ");
    Serial.println(request->url());

    //List all parameters
    int params = request->params();
    for (int i = 0; i < params; i++)
    {
      AsyncWebParameter *p = request->getParam(i);
      if (p->isFile())
      { //p->isPost() is also true
        Serial.printf("FILE[%s]: %s, size: %u\n", p->name().c_str(), p->value().c_str(), p->size());
      }
      else if (p->isPost())
      {
        Serial.printf("POST[%s]: %s\n", p->name().c_str(), p->value().c_str());
      }
      else
      {
        Serial.printf("GET[%s]: %s\n", p->name().c_str(), p->value().c_str());
      }
    }

    if (LittleFS.exists(request->url()) == false)
    {
      request->send(404, "text/plain", "Not found");
      return;
    }

    if (request->url() == "/" || request->url() == "/index.html")
    {
      request->send(LittleFS, "/index.html", String(), false, processor);
      return;
    }

    if (getValue(request->url(), '/', 1) == "css")
    {
      request->send(LittleFS, request->url(), "text/css");
      return;
    }
  }
};

void SevenSeg_Output()
{
  disp_GOF.setBrightness(activeParty == GOF ? 7 : 0);
  disp_MIL.setBrightness(activeParty == MILIZ ? 7 : 0);
  disp_KGG.setBrightness(activeParty == KGG ? 7 : 0);

  disp_GOF.showNumberDecEx(Count_GOF / 20, GOF_dot, true, 4, 0);
  disp_MIL.showNumberDecEx(Count_MILIZ / 20, MIL_dot, true, 4, 0);
  disp_KGG.showNumberDecEx(Count_KGG / 20, KGG_dot, true, 4, 0);
}

void ticker_callback()
{
  switch (activeParty)
  {
  case GOF:
    Count_GOF++;
    GOF_dot = GOF_dot == 0x80 || GOF_dot == 0x00 ? 0x10 : GOF_dot << 1;
    MIL_dot = 0;
    KGG_dot = 0;
    break;

  case MILIZ:
    Count_MILIZ++;
    MIL_dot = MIL_dot == 0x80 || MIL_dot == 0x00 ? 0x10 : MIL_dot << 1;
    GOF_dot = 0;
    KGG_dot = 0;
    break;

  case KGG:
    Count_KGG++;
    KGG_dot = KGG_dot == 0x80 || KGG_dot == 0x00 ? 0x10 : KGG_dot << 1;
    GOF_dot = 0;
    MIL_dot = 0;
    break;

  default:
    break;
  }
}

void serialOut_callback()
{

  static uint32_t SerialPrintCount = 0;

  if (SerialPrintCount % 10 == 0)
  {
    Serial.println("|    GOF    |   MILIZ   |    KGG    |");
  }
  Serial.printf("  %9d   %9d   %9d\n", Count_GOF, Count_MILIZ, Count_KGG);
  SerialPrintCount++;
}

void inputGetter_callback()
{
  activeParty = NONE;

  if (digitalRead(DIO_GOF_TRG) + digitalRead(DIO_MIL_TRG) + digitalRead(DIO_KGG_TRG) < 2)
  {
    Serial.println("ERROR: More than one Flag active");
    return;
  }

  if (digitalRead(DIO_GOF_TRG) == LOW)
    activeParty = GOF;

  if (digitalRead(DIO_KGG_TRG) == LOW)
    activeParty = KGG;

  if (digitalRead(DIO_MIL_TRG) == LOW)
    activeParty = MILIZ;
}

void setup()
{
  pinMode(DIO_GOF_TRG, INPUT_PULLUP);
  pinMode(DIO_MIL_TRG, INPUT_PULLUP);
  pinMode(DIO_KGG_TRG, INPUT_PULLUP);

  Serial.begin(9600);

  LittleFS.begin();

  WiFi.softAP("Dark Emergency CTF Timer");
  dnsServer.start(53, "*", WiFi.softAPIP());

  server.addHandler(new CaptiveRequestHandler()).setFilter(ON_AP_FILTER); //only when requested from AP
  server.begin();

  PartyTicker.start();
  SerialOutputTicker.start();
  InputGetterTicker.start();
}

void loop()
{
  PartyTicker.update();
  SerialOutputTicker.update();
  InputGetterTicker.update();

  dnsServer.processNextRequest();
  SevenSeg_Output();

  while (Serial.available() > 0)
  {
    char input = Serial.read();

    switch (input)
    {
    case 'n':
      activeParty = NONE;
      break;

    case 'g':
      activeParty = GOF;
      break;

    case 'k':
      activeParty = KGG;
      break;

    case 'm':
      activeParty = MILIZ;
      break;

    case 'x':
      SerialOutputTicker.stop();
      Serial.println("SerialOutputTicker.stop()");
      break;

    case 'y':
      SerialOutputTicker.resume();
      Serial.println("SerialOutputTicker.resume()");
      break;

    default:
      break;
    }
  }
}